library ac_dart.addon_routebuilder;

import 'acdart_shelf_middleware.dart';
import 'package:shelf/shelf.dart' as shelf;
import 'package:shelf_route/shelf_route.dart' as route;
import 'package:atlassian_connect_host/atlassian_connect_host.dart';
import 'dart:async';
import 'package:logging/logging.dart';
import 'dart:convert';
import 'package:atlassian_connect_host/host_event_model.dart';
import 'package:path/path.dart' as path;
import 'package:atlassian_connect_shelf/src/util.dart';

final Logger _log = new Logger('AcDartAddonRouteBuilder');

typedef dynamic ACHostServiceRouteHandler(AcHostService acHostService, 
    shelf.Request request);

void _requestLogger(String msg, bool isError) {
  if (isError) {
    _log.warning(msg);
  }
  else {
    _log.finest(msg);
  }
}

final String _assetDir = pathRelativeToAppRoot('asset');

/**
 * A builder for creating routes using the Shelf Route library.
 */
class AcDartMainRouteBuilder extends AcDartRouteBuilder {
  final String _assetPath;
  
  AcDartMainRouteBuilder._internal(AcHostService hostService, AddonHost addonHost, 
      this._assetPath, List<shelf.Middleware> middleware)
      :super._internal(hostService, addonHost, middleware);
  
  AcDartMainRouteBuilder(AcHostService hostService, AddonHost addonHost,
      {bool logRequests: true, String assetPath}) 
      : this._internal(hostService, addonHost, 
          (assetPath != null ? assetPath : _assetDir), 
          (logRequests ? [shelf.logRequests(logger: _requestLogger)] : []));
  
  shelf.Handler get handler => _mainHandler;
  
  /// adds a route to handle the installation lifecycle event. You can override
  /// the [installationPath] if you desire but it is discouraged
  void addInstallationRoute([String installationPath = '/installed']) {
    _mainRouter.addRoute(_installationHandler, path: installationPath, 
        method: 'POST');
  }
  
  /// adds a route to serve the `atlassian-connect.json` file from the standard
  /// location. You can provide additional variables that you can use in the
  /// moustache template.
  void addDescriptorTemplateRoute({ Map<String, String> additionalVariables: const {} }) {
    _mainRouter.addRoute(_descriptorTemplateHander(additionalVariables),
        path: '/atlassian-connect.json', method: 'GET');
  }
  
  
  Future<shelf.Response> _installationHandler(shelf.Request req) {
    _log.fine('received installation request ${req.requestedUri}');

    Future<InstallationEvent> installationEvent = _readJsonBody(req).then(
        (jsonMap) => new InstallationEvent.fromJson(jsonMap));

    Future<AcHostRegistration> registration = installationEvent.then((i) =>
        _hostService.registerHost(i));

    return registration.then((r) {
      return new shelf.Response.ok("installed");
    }, onError: (e, stackTrace) {
      _log.warning('Registration failed', e, stackTrace);
      // TODO: more specific code and message. e.g 404, 401, 403
      return new shelf.Response(400, body: 'Registration failed');
    });
  }
  
  shelf.Handler _descriptorTemplateHander(Map<String, String> additionalVariables) {
    final templateVariables = {
      "addonKey": _addonHost.key,
      "baseUrl": _addonHost.baseUrl.toString()
    }..addAll(additionalVariables);
    
    return moustacheFile(path.join(_assetPath, 'atlassian-connect.json'),
        templateVariables: templateVariables);
  }
  
}

class AcDartRouteBuilder {
  final AcHostService _hostService;
  final route.Router _mainRouter = route.router();
  shelf.Handler _mainHandler;
  final AddonHost _addonHost;
  
  AcDartRouteBuilder._internal(this._hostService, this._addonHost, 
      [ List<shelf.Middleware> middleware = const [] ]) {
    var stack = const shelf.Stack();
    
    stack.addMiddleware(shelf.logRequests());
    
    middleware.forEach((m) => stack = stack.addMiddleware(m));
   
    this._mainHandler = stack.addHandler(_mainRouter.handler);
  }
  
  
  shelf.Handler get handler => _mainHandler;
  
  /// Adds a route that is authenticated with the product host Jwt mechanism
  void addJwtRoute(shelf.Handler handler, String path, { String method }) {
    _mainRouter.addRoute(const shelf.Stack()
      .addMiddleware(productHostAuthenticator(_hostService))
      .addHandler(handler), path: path, method: method);
  }

  /// Adds a route that is authenticated with the product host Jwt mechanism
  /// in which you wish to communicate with the host product
  void addHostServiceJwtRoute(ACHostServiceRouteHandler handler, String path, 
                              { String method }) {
    addJwtRoute(_hostSvcHandler(handler), path, method: method);
  }
  
  /// Adds a route that is authenticated with the addon session token mechanism
  void addSessionTokenRoute(shelf.Handler handler, String path, 
                            { String method }) {
    _mainRouter.addRoute(const shelf.Stack()
      .addMiddleware(sessionTokenAuthenticator(_hostService))
      .addHandler(handler), path: path, method: method);
  }

  /// Adds a route that is authenticated with the addon session token mechanism
  /// in which you wish to communicate with the host product
  void addHostServiceSessionTokenRoute(ACHostServiceRouteHandler handler, 
                                       String path, { String method }) {
    addSessionTokenRoute(_hostSvcHandler(handler), path, method: method);
  }

  /// Adds a route that is public (not authenticated)
  void addUnauthenticatedRoute(shelf.Handler handler, String path, 
                               { String method }) {
    _mainRouter.addRoute(handler, path: path, method: method);
  }

  /// Adds a route that is public (not authenticated) in which you want to 
  /// communicate with the host in an unauthenticated manner
  void addUnauthenticatedHostServiceRoute(ACHostServiceRouteHandler handler, 
                                       String path, { String method }) {
    addUnauthenticatedRoute(_hostSvcHandler(handler), path, method: method);
  }
  
  /// creates a child route builder at the given path
  AcDartRouteBuilder addUnauthenticatedChildBuilder(String path) {
    final childBuilder = new AcDartRouteBuilder._internal(_hostService, 
        _addonHost);
    addUnauthenticatedRoute(childBuilder.handler, path);
    return childBuilder;
  }

  /// creates a child route builder at the given path where all child routes
  /// are authenticated with the product host Jwt mechanism
  AcDartRouteBuilder addJwtChildBuilder(String path) {
    final childBuilder = new AcDartRouteBuilder._internal(_hostService, 
        _addonHost);
    addJwtRoute(childBuilder.handler, path);
    return childBuilder;
  }
  
  /// creates a child route builder at the given path where all child routes
  /// are authenticated with the addon session token mechanism
  AcDartRouteBuilder addSessionTokenChildBuilder(String path) {
    final childBuilder = new AcDartRouteBuilder._internal(_hostService,
        _addonHost);
    addSessionTokenRoute(childBuilder.handler, path);
    return childBuilder;
  }
  
  shelf.Handler _hostSvcHandler(ACHostServiceRouteHandler handler) =>
      (shelf.Request request) {
    return handler(_hostService, request);
  };
  
  Future<Map<String, Object>> _readJsonBody(shelf.Request req) {
    return req.readAsString().then((String s) => JSON.decode(s));
  }

}



